/*
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#ifndef _INET_H_
#define _INET_H_

/* XXX These are machine/compiler dependent */
typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;

#define IPPROTOCOL_ICMP		1
#define IPPROTOCOL_IGMP		2
#define IPPROTOCOL_TCP		6
#define IPPROTOCOL_UDP		17
#define IP_DF 0x4000

/* TCP Flags */
#define TCPFLAGS_FIN	0x01
#define TCPFLAGS_SYN	0x02
#define TCPFLAGS_RST	0x04
#define TCPFLAGS_PSH	0x08
#define TCPFLAGS_ACK	0x10
#define TCPFLAGS_URG	0x20
#define TCPFLAGS_ECN_ECHO	0x40
#define TCPFLAGS_CWR	0x80

#define TCPOPT_EOL    0 
#define TCPOLEN_EOL   1
#define TCPOPT_NOP    1
#define TCPOLEN_NOP   1
#define TCPOPT_MAXSEG   2
#define TCPOLEN_MAXSEG   4
#define TCPOPT_WINDOW   3
#define TCPOLEN_WINDOW   3
#define TCPOPT_SACK_PERMITTED 4   
#define TCPOLEN_SACK_PERMITTED 2
#define TCPOPT_SACK   5   
#define TCPOPT_TIMESTAMP  8
#define TCPOLEN_TIMESTAMP    10

struct IpHeader {
  uint8		ip_vhl;	/* version (4bits) & header length (4 bits) */
  uint8		ip_tos;	/* type of service */
  uint16	ip_len; /* length of IP datagram */
  uint16	ip_id;	/* identification (for frags) */ 
  uint16	ip_off; /* offset (within a fragment) and flags (3 bits) */
  uint8		ip_ttl; /* time to live */
  uint8		ip_p;	/* protocol number */
  uint16	ip_xsum; /* checksum */
  uint32	ip_src; /* source address */
  uint32	ip_dst; /* destination address */
};

/* Pseudo header for doing TCP checksum calculation */
struct PseudoIpHeader {
  uint32	filler[2];
  uint8		zero;
  uint8		ip_p;
  uint16	ip_len;
  uint32	ip_src;
  uint32	ip_dst;
};

struct TcpHeader {
  uint16	tcp_sport;	/* source port */
  uint16	tcp_dport;	/* destination port */
  uint32	tcp_seq;	/* sequence number */
  uint32	tcp_ack;	/* acknoledgement number */
  uint8		tcp_hl;		/* header length (4 bits) */
  uint8		tcp_flags;	/* flags */
  uint16	tcp_win;	/* advertized window size */
  uint16	tcp_xsum;	/* checksum */
  uint16	tcp_urp;	/* urgent pointer */

};

struct IPPacket {
  struct IpHeader ip;
  struct TcpHeader tcp;
};


char *InetAddress(uint32 addr);
uint16 InetChecksum(uint16 *addr, uint16 len);
void PrintTcpPacket(struct IPPacket *p);
void WriteIPPacket(struct IPPacket *p,
		    uint32 src, uint32 dst, uint16 sport, uint16 dport,
		    uint32 seq, uint32 ack, uint8 flags, uint16 win,
		    uint16 urp, uint16 datalen, uint16 optlen, uint8 iptos, uint8 u4tf); 
void ReadIPPacket(struct IPPacket *p, uint32 *src, uint32 *dst, 
                   uint16 *sport, uint16 *dport, uint32 *seq, uint32 *ack, 
									 uint8 *flags, uint16 *win, uint16 *urp, uint16 *datalen, 
									 uint16 *optlen);
void StorePacket (struct IPPacket *p); 

#endif /* _INET_H_ */
