/* 
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#include "base.h"
#include "inet.h"
#include "session.h"
#include "support.h"
#include "tbit.h"
#include "history.h"
#include "reno.h"

#define MAXRXMT 20  
#define MAXTO 20  

extern struct TcpSession session;
extern struct History history[];

static char *renoVersionNames [] = {
	"Uncategorized", 
	"Tahoe", 
	"TahoeNoFR", 
	"Reno", 
	"NewReno",
	"AggresiveReno", 
	"AggresiveTahoeNoFR",
	"RenoPlus",
	"RenoNS"};

static uint32 renoDrop[2]; 
static int droppedForRenoTest; 

void RenoTest (uint32 sourceIpAddress, uint16 sourcePort, uint32 targetIpAddress, uint16 targetPort, int mss) 
{
	int optlen; 
	char *opt; 

	optlen = 4 ;
	if ((opt=(char *)calloc(sizeof(uint8), optlen)) == NULL)	{
		perror("ERROR: Could not allocate opt:");
		Quit(ERR_MEM_ALLOC);
	}

	/* mss option */
	opt[0] = (uint8)TCPOPT_MAXSEG ; 
	opt[1] = (uint8)TCPOLEN_MAXSEG ; 
	*((uint16 *)((char *)opt+2)) = htons(mss);

	if (EstablishSession(sourceIpAddress,sourcePort,targetIpAddress,targetPort,mss,optlen,opt,5,100,0) == 0) {
		printf("ERROR: Couldn't establish session\n");
		Quit(NO_SESSION_ESTABLISH);
	}
	SendRequest(session.filename, (void *)RenoackData); 
	rcvData (RenoackData);
}

void RenoackData (struct IPPacket *p) 
{
	uint32 seq = history[session.hsz-1].seqno;
	uint8  fin = history[session.hsz-1].fin;
	uint16 datalen = history[session.hsz-1].dlen;
	char  *dataptr = history[session.hsz-1].data;
	int i ; 
	struct IPPacket ackpkt ; 
	int dropForRenoTest = 0 ; 

	if (session.debug) {
		printf("\n====================\n");
		printf ("datalen = %d\n\n", (int)datalen);
		if (datalen > 0) {
			for (i = 0 ; i < (int)datalen ; i++) {
				printf ("%c", isprint(dataptr[i])?dataptr[i]:' '); 
			}
		}
		printf("\n====================\n");
	}

	if (datalen > session.mss) {
		printf ("ERROR: mss=%d datalen=%d\n", session.mss, datalen);
		Quit(MSS_ERR);
	}

	if ((seq-session.irs+datalen) > session.mss*session.maxpkts) {
		printf ("ERROR: buffer overflow: %u %d %u %d %d\n", 
						seq, datalen, session.irs, seq+datalen-session.irs, session.mss*session.maxpkts);
		Quit(BUFFER_OVERFLOW); 
	}

	if (datalen > 0) {
		session.totDataPktsRcvd ++ ;
		if (session.verbose) {
			printf ("r %f %d %d\n", GetTime() - session.epochTime, seq-session.irs, seq-session.irs+datalen);
		}
	}

	if(session.maxseqseen < seq+datalen-1) {
		session.maxseqseen = seq +datalen-1; 
	}
	else {
		if (datalen > 0) {
			printf ("##### oos packet %f %d %d max=%d\n", GetTime()-session.epochTime, seq-session.irs, 
										 seq-session.irs+datalen, session.maxseqseen-session.irs);
			if ((seq == renoDrop[0]) || (seq == renoDrop[1])) {
				session.totOutofSeq++;
			}
			if (!((seq == renoDrop[0])||(seq == renoDrop[1])) && (session.totOutofSeq < 2)) {
				if (reordered(p)) {
					printf ("ERROR: Reordering before test completion\n");
					Quit(UNWANTED_PKT_REORDER);
				}
				else {
					printf ("ERROR: unwanted packet drop before test completion\n");
					Quit(UNWANTED_PKT_DROP); 
				}
			}
		}
	}

	/*
	 * drop packets for reno version test? 
	 *
	 */
	if (((droppedForRenoTest==0) && ((seq-session.irs) > 12*session.mss)) ||
		  ((droppedForRenoTest==1) && ((seq-session.irs) > 15*session.mss))) {
		if (droppedForRenoTest >= 2) {
			printf ("Too many drops\n");
			Quit(FAIL); 
		}
		renoDrop[droppedForRenoTest++] = seq ;	
		dropForRenoTest = 1;
		printf ("##### droppacket %f %d %d\n", GetTime()-session.epochTime, seq-session.irs, seq-session.irs+datalen);
	}
	
	if (session.debug) {
		printf ("dlen=%d reno=%d seq=%u rcv_nxt=%u rcv_wnd=%u\n", datalen, dropForRenoTest, 
						 seq, session.rcv_nxt, session.rcv_wnd);
	}
	if ((datalen > 0) && 
			(dropForRenoTest == 0) &&
			(session.rcv_nxt <= seq) && (seq < (session.rcv_nxt+session.rcv_wnd))  &&
			(session.rcv_nxt <= (seq+datalen)) && ((seq+datalen-1) < (session.rcv_nxt + session.rcv_wnd))) {
		int start = seq - session.irs ; 
		int end = start + datalen ; 
		if (session.debug) {
			printf ("rcved = %d-%d\n", start, end);
		}
		for (i = start ; i < end ; i++) {
			session.dataRcvd[i] = 1 ; 
		}
		start = session.rcv_nxt - session.irs ; 
		end = session.mss*session.maxpkts ; 
		for (i = start ; i < end ; i ++) {
			if (session.dataRcvd[i] == 0) {
				break ;
			}
			session.rcv_nxt ++ ;
		}
		if (session.debug) {
			printf ("rcv_nxt = %u\n", (int)session.rcv_nxt);
		}
		if ((droppedForRenoTest == 2) && (session.rcv_nxt >= 6*session.mss+renoDrop[1])) {
			WhichReno ();
			Quit(SUCCESS);
		}
	}

	if (fin) {
		Quit(NOT_ENOUGH_PKTS);
	}
	else {
		if ((datalen > 0) && (dropForRenoTest == 0)) {
			if (session.debug) { 
				printf ("sending ack\n");
			}
			busy_wait(PLOTDIFF);
			SendSessionPacket (&ackpkt, sizeof(struct IPPacket), TCPFLAGS_ACK, 0, 0);
			if (session.verbose) {
				printf ("a %f %d\n", GetTime() - session.epochTime, session.rcv_nxt-session.irs);
			}
		}
	}
}

void WhichReno () 
{
	int rxmt[MAXRXMT] ; 
	int to[MAXTO] ; 
	int i, j ; 
	int numRxmt = 0, numTO = 0 ;	
	int renoversion = Uncategorized ; 
	int firstDropTO = 0;
	int secondDropTO = 0;
	int count = 0; 

	/* 
	 * lets count retransmissions 
	 * using a very dumb algo, but
	 * what the heck ....
	 * note that the first pkt can not
	 * be a retransmission
	 */

	for (i = 1; i < session.hsz; i++) {
		if ((history[i].type == RCVD) && (history[i].dlen > 0)) {
			for (j = 0 ; j < i-1 ; j++) {
				if ( (history[j].type == RCVD) && (history[j].dlen > 0) &&
						 (history[j].seqno <= history[i].seqno) &&
						 (history[j].nextbyte > history[i].seqno)) {
					rxmt[numRxmt++] = i ; 
					if (numRxmt == MAXRXMT) {
						renoversion = Uncategorized ; 
					}
					break ;
				}
			}
		}
	}

	/*
	 * lets now count timeouts 
	 * just as dumb as before 
	 * there can not be a timeout before
	 * the first packet
	 */

	for (i = 1 ; i < session.hsz ; i ++) {
		if ((history[i].type == RCVD) && (history[i].dlen > 0)) {
			int prevDataPkt = -1 ;
			for (j = i-1; j >= 0 ; j --) {
				if ((history[j].type == RCVD) && (history[j].dlen > 0)) {
					prevDataPkt = j ;	
					break ;
				}
			}
			if (prevDataPkt > 0) {
				if ((history[i].timestamp - history[prevDataPkt].timestamp)	> RTT_TO_MULT*(session.rtt+PLOTDIFF)) {
					/* sanity check for rxmt? */
					/* data pkt after TO must be a retransmission */
					int j ;
					int found = 0 ;
					for (j = 0 ; j < numRxmt ; j++) {
						if (rxmt[j] == i) {
							found = 1 ; 
							if (session.verbose) {
								printf ("#### timeout: prev=%d prevtime=%f this=%d thistime=%f timediff=%f thresh=%f\n", 
												 history[prevDataPkt].seqno-session.irs, history[prevDataPkt].timestamp, 
												 history[i].seqno-session.irs, history[i].timestamp, 
												 (history[i].timestamp - history[prevDataPkt].timestamp), 
												 RTT_TO_MULT*(session.rtt+PLOTDIFF));
							}
							break ; 
						}
					}
					if (found == 1) {
						to[numTO++]	 = i ;
						if (numTO == MAXTO) {
							renoversion = Uncategorized ; 
						}
					}
				}
			}
		}
	}

	/*
	 * were there Timeouts for first or second [packet that we dropped?
	 */

	for (i = 0; i < numTO ; i++) {
		if (history[to[i]].seqno == renoDrop[0]) {
			firstDropTO = 1;
		}
		if (history[to[i]].seqno == renoDrop[1]) {
			secondDropTO = 1;
		}
	}

	if ((firstDropTO==0) &&
			(secondDropTO==0) &&
			((rxmt[1]-rxmt[0])==2) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==renoDrop[1]) &&
			((numRxmt==2)||(history[rxmt[2]].seqno > history[rxmt[1]].nextbyte))) {
		count ++; 
		renoversion=NewReno;
	}
	if ((firstDropTO==0) &&
			(secondDropTO==0) &&
			((rxmt[1]-rxmt[0])>2) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==renoDrop[1]) &&
			((numRxmt==2)||(history[rxmt[2]].seqno > history[rxmt[1]].nextbyte))) {
		count ++; 
		renoversion=RenoPlus;
	}
	if ((firstDropTO==0) &&
			(secondDropTO==0) &&
			(numRxmt>3) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==history[rxmt[0]].nextbyte) &&
			(history[rxmt[2]].seqno==renoDrop[1]) &&
			(history[rxmt[3]].seqno==history[rxmt[2]].nextbyte)) {
		count ++; 
		renoversion=AggresiveReno;
	}
	/*
	if ((firstDropTO==0) &&
			(secondDropTO==0) &&
			((rxmt[1]-rxmt[0])==2) &&
			(numRxmt>2) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==renoDrop[1]) &&
			(history[rxmt[2]].seqno==history[rxmt[1]].nextbyte)) {
		count ++; 
		renoversion=RenoNS;
	}
	*/
	if ((firstDropTO==0) &&
			(secondDropTO==1) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==renoDrop[1]) &&
			((numRxmt==2)||(history[rxmt[2]].seqno > history[rxmt[1]].nextbyte))) {
		count ++; 
		renoversion=Reno;
	}
	if ((firstDropTO==0) &&
			(secondDropTO==0) &&
			(numRxmt>2) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==renoDrop[1]) &&
			(history[rxmt[2]].seqno==history[rxmt[1]].nextbyte)) {
		count ++; 
		renoversion=Tahoe;
	}
	if ((firstDropTO==1) &&
			(secondDropTO==0) &&
			(numRxmt>2) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==renoDrop[1]) &&
			(history[rxmt[2]].seqno==history[rxmt[1]].nextbyte)) {
		count ++; 
		renoversion=TahoeNoFR;
	}
	if ((firstDropTO==1) &&
			(secondDropTO==0) &&
			(numRxmt>2) &&
			(history[rxmt[0]].seqno==renoDrop[0]) &&
			(history[rxmt[1]].seqno==history[rxmt[0]].nextbyte) &&
			(history[rxmt[2]].seqno==renoDrop[1])) {
		count ++; 
		renoversion=AggresiveTahoeNoFR;
	}
	printf ("#### rx=%d to=%d version=%s rtt_unr=%d reorder=%d count=%d\n", numRxmt, numTO, 
					 renoVersionNames[renoversion], session.rtt_unreliable, session.num_reordered, count);
}
