/* 
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#include "base.h"
#include "inet.h"
#include "session.h"
#include "capture.h"
#include "support.h"
#include "tbit.h"
#include "history.h"
#include "sack_rcvr.h"

extern struct TcpSession session;
extern struct History history[];

void SackRcvrTest (uint32 sourceAddress, uint16 sourcePort, uint32 targetAddress, uint16 targetPort, int mss) 
{
	int optlen ; 
	char *opt	; 

	optlen = 8 ;
	if ((opt=(char *)calloc(sizeof(uint8), optlen)) == NULL)	{
		perror("ERROR: Could not allocate opt:");
		Quit(ERR_MEM_ALLOC);
	}
	/* mss option */
	opt[0] = (uint8)TCPOPT_MAXSEG ; 
	opt[1] = (uint8)TCPOLEN_MAXSEG ; 
	*((uint16 *)((char *)opt+2)) = htons(mss);
	/* align 4-byte boundries */
	opt[4] = (uint8)TCPOPT_NOP ;
	opt[5] = (uint8)TCPOPT_NOP ;
	/* sack option */
	opt[6] = (uint8)TCPOPT_SACK_PERMITTED ;
	opt[7] = (uint8)TCPOLEN_SACK_PERMITTED ;

	if (EstablishSession(sourceAddress,sourcePort,targetAddress,targetPort,mss,optlen,opt,5,5,0) == 0) {
		printf("ERROR: Couldn't establish session\n");
		Quit(NO_SESSION_ESTABLISH);
	}
	if (!isSack ()) {
		printf("ERROR: remote host does not permit SACK\n");
		Quit(NO_SACK);
	}
	SendBrokenRequest();
	checkSackRcvr();
	Quit(SUCCESS);
}

int isSack () 
{
	int i,j ; 
	for (i = 1; i < session.hsz; i++) {
		if ((history[i].type == RCVD) && (history[i].ack) && (history[i].syn)) {
			j = 0 ;
			while (j < history[i].optlen) {
				switch ((unsigned char)history[i].opt[j]) {
					case TCPOPT_EOL:
						j = j + TCPOLEN_EOL;
						break; 
					case TCPOPT_NOP:
						j = j + TCPOLEN_NOP ;
						break;
					case TCPOPT_SACK_PERMITTED:
						return 1;
						break ; 
					default: 
						if ((uint8)history[i].opt[j+1] > 0) {
							j = j + (uint8)history[i].opt[j+1] ;
						} 
						else {
				 			Quit(BAD_OPT_LEN); 
						}
						break ;
				}
			} 
		}
	}
	return 0;
}

void SendBrokenRequest() 
{
	struct IPPacket *p, *datapkt;
	struct PacketInfo pi;
	int i ;
	uint16 lastSeqSent = session.snd_nxt; 
	double startTime;
	char *dataptr ; 
	char data[MAXREQUESTLEN];
	int datalen;
	int ipsz; 

	p = NULL; 
	datapkt = NULL;

	/*
	 * we will send data packets 1 and 3. 
	 * the ack for 1 should have no sack
	 * ack for 3 should have a 
	 * sack blocks indicating a hole at 2. 
	 * each packet will be 1 byte long. 
	 * we will wait for 2 seconds after each packet is sent to 
	 * receive an ack, otherwise we will quit.
	 */

	datalen = PrepareRequest (data, session.filename);

	ipsz = sizeof(struct IPPacket) + 1; 
	for (i = 0 ; i < 14 ; i++) {
		if (((char *)datapkt=(char *)calloc(ipsz, 1)) == NULL) {
			perror("ERROR: no space for request packet:");
			Quit(ERR_MEM_ALLOC) ; 
		}
		dataptr = (char *)datapkt + sizeof(struct IPPacket);
		memcpy((void *)dataptr, (void *)data+i, 1);
		SendSessionPacket(datapkt, ipsz, TCPFLAGS_PSH | TCPFLAGS_ACK, 0, 0);
		startTime = GetTime();
		while (1) {
			/* Check if we have received any packets */
			if ((p =(struct IPPacket *)CaptureGetPacket(&pi)) != NULL) {
				/*
			 	 * packet that we sent?
			 	 */
				if (INSESSION(p,session.src,session.sport,session.dst,session.dport) &&
				    (p->tcp.tcp_flags == (TCPFLAGS_PSH | TCPFLAGS_ACK)) &&
				    (ntohl(p->tcp.tcp_seq) > lastSeqSent) &&
				    (ntohl(p->tcp.tcp_ack) <= session.rcv_nxt)) {
					lastSeqSent = ntohl(p->tcp.tcp_seq);
					if (session.debug) {
						printf("saw the packet we sent: xmit %d\n", i);
						PrintTcpPacket(p);
					}
					free (datapkt);
					StorePacket(p);
					session.snd_nxt ++;
					session.totSeenSent ++ ;
					continue ;
				} 
				/*
			 	 * from them? 
			 	 */ 
				if (INSESSION(p,session.dst,session.dport,session.src,session.sport) &&
					  (p->tcp.tcp_flags & TCPFLAGS_ACK) &&
					  (ntohl(p->tcp.tcp_seq) == session.rcv_nxt) &&
					  (ntohl(p->tcp.tcp_ack) >= session.snd_una)) {
					if (ntohl(p->tcp.tcp_ack) > session.snd_una) {
						session.snd_una  = ntohl(p->tcp.tcp_ack);
					}
					if (session.debug) {
						PrintTcpPacket(p);
					}
					StorePacket(p);
					session.totRcvd ++;
					break ;
				}
				/* 
			 	 * otherwise, this is a bad packet
			 	 * we must quit
			 	 */
			 	processBadPacket(p);
			}
			if (GetTime()-startTime>=REXMITDELAY) {
				printf ("ERROR: no response\n");
				Quit(NO_CONNECTION);
			}
		}
		/* now, bump up the snd_nxt to create a false loss
		 * and skip an iteration. 
		 * is skipping iteration necessary?
		 */
		session.snd_nxt ++ ;
		i++;
	}
}

void checkSackRcvr() 
{
	/*
	 * we check if there is a "normal" ack for the first packet
	 * and a "sack" for the second packet. 
	 * sack for second packet should have appropriate sack blocks
	 */

	int i, j, k;
	int count = 0;
	uint32 sb[MAXSB];
	int numsb = -1; 
	int maxsack;

	for (i = 0 ; i < session.hsz; i++) {
		if ((history[i].type == RCVD) && (history[i].ack == 1) && 
				(history[i].syn == 0) && (history[i].dlen == 0)) {
			count ++; 
			if (count==1) {
				if (!(history[i].ackno-session.iss) == 2) {
					printf ("Incorrect behavior on first ack\n");
					Quit(SUCCESS);
				}
			}
			if (count> 1) {
				if (history[i].optlen == 0) {
					printf ("No sack block in ack %d %d\n", i, session.hsz);
					Quit(SUCCESS);
				}
				j = 0;
				while (j < history[i].optlen) {
					switch ((unsigned char)history[i].opt[j]) {
						default: 
							if ((uint8)history[i].opt[j+1] > 0) {
								j = j + (uint8)history[i].opt[j+1] ;
							}
							else {
								Quit (BAD_OPT_LEN);
							}
							break;
						case TCPOPT_EOL:
							j = j + TCPOLEN_EOL;
							break;
						case TCPOPT_NOP:
							j = j + TCPOLEN_NOP ;
							break;
						case TCPOPT_SACK:
							GetSB (history[i].opt, history[i].optlen, j, sb, &numsb);
							maxsack = 2*count - 1;
							for (k = 0 ; k < 2*numsb ; k+=2) {
								if (!((sb[k]-session.iss == maxsack) && (sb[k+1]-session.iss == maxsack+1))) {
									printf ("1:Error in sack block %d\n", count -1);
									Quit(SUCCESS);
								}
								maxsack -= 2;
							}
							j = j + (uint8)history[i].opt[j+1] ;
							break;
					}
				}
			}
		}
	}
	if (count != 7) {
		printf ("Error in sack sequence\n");
		Quit(SUCCESS);
	}
	else {
		printf ("sack okay\n");
		Quit(SUCCESS);
	}
}

void GetSB (void *opt, int optlen, int ptr, uint32 *sb, int *numsb)
{
	int len = ((uint8*)opt)[ptr+1] ;
	int i;
	if (len > 2) {
		*numsb = (len-2)/8;
	} else {
		Quit (BAD_OPT_LEN);
	}
	for (i = 0 ; i < 2*(*numsb); i++) {
		memcpy (&sb[i], ((char *)opt)+ptr+2+i*sizeof(uint32), sizeof(uint32));	
		sb[i] = ntohl (sb[i]);
	}
}
