/* 
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#include "base.h"
#include "inet.h"
#include "session.h"
#include "support.h"
#include "tbit.h"
#include "history.h"
#include "sack_rcvr.h"
#include "sack_sndr_3p.h"

#define MAXRXMT 20  
#define SB 20 

extern struct TcpSession session;
extern struct History history[];

static uint32 sackDrop[3]; 
static int droppedForSackTest; 

void SackSndr3PTest (uint32 sourceIpAddress, uint16 sourcePort, uint32 targetIpAddress, uint16 targetPort, int mss) 
{
	int optlen; 
	char *opt; 

	optlen = 8;
	if ((opt=(char *)calloc(sizeof(uint8), optlen)) == NULL)	{
		perror("ERROR: Could not allocate opt:");
		Quit(ERR_MEM_ALLOC);
	}
	/* mss option */
	opt[0] = (uint8)TCPOPT_MAXSEG ; 
	opt[1] = (uint8)TCPOLEN_MAXSEG ; 
	*((uint16 *)((char *)opt+2)) = htons(mss);
	/* align 4-byte boundries */
	opt[4] = (uint8)TCPOPT_NOP ;
	opt[5] = (uint8)TCPOPT_NOP ;
	/* sack option */
	opt[6] = (uint8)TCPOPT_SACK_PERMITTED ;
	opt[7] = (uint8)TCPOLEN_SACK_PERMITTED ;
	if (EstablishSession(sourceIpAddress,sourcePort,targetIpAddress,targetPort,mss,optlen,opt,16,100,0) == 0) {
		printf("ERROR: Couldn't establish session\n");
		Quit(NO_SESSION_ESTABLISH);
	}
	
	if (!isSack ()) {
		printf ("#### remote host does not support sack\n");
		Quit(NO_SACK);
	}
	SendRequest(session.filename, (void *)Sack3PData); 
	rcvData (Sack3PData);
}

void Sack3PData (struct IPPacket *p) 
{
	uint32 src;
	uint32 dst;
	uint16 sport;
	uint16 dport;
	uint32 seq;
	uint32 ack;
	uint8  flags;
	uint16 win;
	uint16 urp;
	uint16 datalen;
	uint16 optlen;
	char  *dataptr ;
	int i ; 
	struct IPPacket *ackpkt ; 
	int dropForSackTest = 0 ; 
	char *opt; 
	int olen;

	opt = NULL;
	ReadIPPacket (p, &src, &dst, &sport, &dport, &seq, &ack, &flags, &win,
								 &urp, &datalen, &optlen);
	if (session.debug) {
		printf ("datalen = %d\n\n", (int)datalen);
		if (datalen > 0) {
			dataptr = (char *)p + (int)sizeof(struct IPPacket) + (int)optlen ;		
			for (i = 0 ; i < (int)datalen ; i++) {
				//printf ("%c", isprint(dataptr[i])?dataptr[i]:' '); 
			}
		}
		printf("\n====================\n");
	}

	if (datalen > session.mss) {
		printf ("ERROR: mss=%d datalen=%d\n", session.mss, datalen);
		Quit(MSS_ERR);
	}
	if ((seq+datalen-session.irs) > session.mss*session.maxpkts) {
		printf ("ERROR: buffer overflow: %u %d %u %d %d\n", 
						seq, datalen, session.irs, seq+datalen-session.irs, session.mss*session.maxpkts);
		Quit(BUFFER_OVERFLOW); 
	}

	if (datalen > 0) {
		session.totDataPktsRcvd ++ ;
		if (session.totDataPktsRcvd >= session.maxpkts) {
			if (session.totOutofSeq < 3) {
				Quit(NOT_ENOUGH_PKTS);
			} else {
				Sack3PCheck() ;
				Quit(SUCCESS); 
			}
		}
		if (session.verbose) {
			printf ("r %f %d %d %d\n", GetTime() - session.epochTime, seq-session.irs, seq-session.irs+datalen,
			session.totDataPktsRcvd);
		}
	}
	if(session.maxseqseen <= seq+datalen-1) {
		session.maxseqseen = seq +datalen-1; 
	}
	else {
		if (datalen > 0) {
			printf ("##### rexmit packet %f %d %d \n", GetTime()-session.epochTime, seq-session.irs, seq-session.irs+datalen);
			session.totOutofSeq ++ ;
			if ((session.totOutofSeq == 3) && (seq == sackDrop[2])) {
				Sack3PCheck();
				Quit(SUCCESS);
			}
			if (session.totOutofSeq == 3) {
				Sack3PCheck();
				Quit(SUCCESS);
			}
			if ((seq != sackDrop[0]) && (seq != sackDrop[1]) && (seq != sackDrop[2]) && (session.totOutofSeq < 3)) {
				if (reordered(p)) {
					printf ("ERROR: Reordering before test completion\n");
					Quit(UNWANTED_PKT_REORDER);
				}
				else {
					printf ("ERROR: unwanted packet drop befoer test completion\n");
					Quit(UNWANTED_PKT_DROP); 
				}
			}
		}
	}

	/*
	 * drop packets for sack test?  
	 */
	if ( (session.totDataPktsRcvd == 15) || 
			 (session.totDataPktsRcvd == 17) ||
			 (session.totDataPktsRcvd == 19)) { 
		sackDrop[droppedForSackTest++] = seq ;	
		dropForSackTest = 1;
		printf ("##### droppacket %f %d %d\n", GetTime()-session.epochTime, seq-session.irs, seq-session.irs+datalen);
	}
	
	/* from TCP/IP vol. 2, p 808 */
	if (session.debug) {
		printf ("dlen=%d sack=%d seq=%u rcv_nxt=%u rcv_wnd=%u\n", datalen, dropForSackTest, 
						 seq, session.rcv_nxt, session.rcv_wnd);
	}
	if ((datalen > 0) && 
			(dropForSackTest == 0) &&
			(session.rcv_nxt <= seq) && (seq < (session.rcv_nxt+session.rcv_wnd))  &&
			(session.rcv_nxt <= (seq+datalen)) && ((seq+datalen-1) < (session.rcv_nxt + session.rcv_wnd))) {
		int start = seq - session.irs ; 
		int end = start + datalen ; 
		if (session.debug) {
			printf ("rcvd = %d-%d\n", start, end);
		}
		for (i = start ; i < end ; i++) {
			session.dataRcvd[i] = 1 ; 
		}
		start = session.rcv_nxt - session.irs ; 
		end = session.mss*session.maxpkts ; 
		for (i = start ; i < end ; i ++) {
			if (session.dataRcvd[i] == 0) {
				break ;
			}
			session.rcv_nxt ++ ;
		}
		if (session.debug) {
			printf ("rcv_nxt = %u\n", (int)session.rcv_nxt);
		}
	}

	if (flags & TCPFLAGS_FIN) {
		if (session.totOutofSeq == 3) {
			Sack3PCheck () ;
			Quit(SUCCESS); 
		}
	}
	if ((datalen > 0) && (dropForSackTest == 0)) {
		if (session.debug) { 
			printf ("sending ack\n");
		}
		busy_wait(PLOTDIFF);
		olen= 0;
		if (session.maxseqseen > session.rcv_nxt) {
			opt = FillSack3POpt(&olen);
		}
	  /* allocate the ack packet */
  	if ((ackpkt = (struct IPPacket *)calloc(1, sizeof(struct IPPacket)+olen)) == NULL) {
   		perror("ERROR: Could not allocate SYN packet:") ;
   		Quit(ERR_MEM_ALLOC) ;
  	}
  	/* fill in the options */
  	if (olen > 0) {
   		memcpy((char *)ackpkt+sizeof(struct IPPacket), opt, olen);
  	}
		SendSessionPacket (ackpkt, sizeof(struct IPPacket)+olen, TCPFLAGS_ACK, olen, 0);
		free (ackpkt);
		free(opt);
		if (session.verbose) {
			printf ("a %f %d\n", GetTime() - session.epochTime, session.rcv_nxt-session.irs);
		}
	}
}

char *FillSack3POpt(int *optlen) 
{
	uint32 sb[2*SB];
	int start, end;
	int inhole = 0;
	char *opt;
	int edges = 0; 

	start = session.maxseqseen - session.irs;
	end = start;
	while (start >= 0) {
		switch (session.dataRcvd[start]) {
			case 1: 
				if (inhole) {
					end = start;
					inhole = 0;
				}
				break;
			case 0: 
				if (!inhole) {
					inhole = 1;
					if (edges == 2*SB) {
						printf ("Too many holes\n");
						Quit (TOO_MANY_HOLES);
					}
					sb[edges] = start+1+session.irs;
					sb[edges] = htonl(sb[edges]);
					edges++;
					sb[edges] = end+1+session.irs;
					sb[edges] = htonl(sb[edges]);
					edges++;
				}
				break;
			default: 
				printf ("Error in sack block calulcation\n");
				Quit(ERR_IN_SB_CALC);
				break;
		}
		start --;
	}
	
	*optlen = 4+sizeof(uint32)*edges; 
	if ((opt=(char *)calloc(sizeof(uint8), *optlen)) == NULL)	{
		perror("ERROR: Could not allocate opt:");
		Quit(ERR_MEM_ALLOC);
	}
	opt[0] = (uint8)TCPOPT_SACK ; 
	opt[1] = (uint8)(2+4*edges) ;
	memcpy(opt+2, sb, edges*sizeof(uint32));	
	opt[*optlen-2] = (uint8)TCPOPT_NOP ;
	opt[*optlen-1] = (uint8)TCPOPT_NOP ;
	return opt;
}

void Sack3PCheck () 
{
	int sack = 1;	
	int i, j;	
	double start=0; 
	double end=999999999999999;
	uint32 rxmt[MAXRXMT];
	int numRxmt = 0;

	/* check for unnecessary retransmissions */
	for (i = 1; ((i<session.hsz)&&(sack == 1)); i++) {
		if ((history[i].type == RCVD) && (history[i].dlen > 0)) {
			for (j = 0 ; j < i-1 ; j++) {
				if ( (history[j].type == RCVD) && (history[j].dlen > 0) &&
						 (history[j].seqno <= history[i].seqno) &&
						 (history[j].nextbyte > history[i].seqno)) {
					rxmt[numRxmt++] = i ; 
					if ((numRxmt == 1) && (history[i].seqno == sackDrop[0])) {
						start = history[i].timestamp;
					}
					if (numRxmt > 3) {
						sack = 0;
						break;
					}
					if ((numRxmt == 3) && (history[i].seqno == sackDrop[2])) {
						end = history[i].timestamp;
					}
				}
			}
		}
	}
	if ((end - start) > 2*session.rtt) {
		sack =0;
	}
	if (session.verbose) {
		printf ("#### %f %f %d\n", (end - start), 2*session.rtt, numRxmt);
	}
	printf ("#### sack = %d\n", sack);	
}
